/***************************************************************************
               uconfigdb.h  -  http://uconfig.sourceforge.net
               ----------------------------------------------
    begin                : Mon Sep 11 2000
    copyright            : (C) 2000 by Kir Kostchenko
    email                : kir@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU Lesser General Public License as         *
 *  published by the Free Software Foundation; either version 2.1 of the   *
 *  License, or (at your option) any later version.                        *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *  File require Qt 2.1 or newer library. http://www.trolltech.com         *
 *  File documented with doc++.  http://linuxsupportline.com/~doc++        *
 *                                                                         *
 ***************************************************************************/

#ifndef __UCONFIGDB_H
#define __UCONFIGDB_H

#include <qstring.h>
#include <qbuffer.h>
#include <qdatastream.h>
#include <qlist.h>
#include <qvariant.h>

#define __UCDEBUG 1

#ifdef __UCDEBUG
  #include <assert.h>
#endif

class UConfig;
class UCEntriesIterator;

  /**   It's internal UConfig namespace.
   * I can not understand egcs bug, when
   * declare this classes inside UConfig class.
   * But only that way I can declare internal
   * template classes for UConfig private
   * operations.
   *  Working with defines is awful...
   *
   *    Qt does not declare internal namespace.
   * They do classes named QGCache, QGDict and so
   * in global namespace. I think it's not right.
   */

namespace UConfigPrivate {

   /** Internal - address from the beginning of file. */
   typedef Q_UINT32 FAddr;  // File address.

      // Parent header type.
   enum ParentType { LeftChild,
                     RightChild,
                     RootChild,
                     FoundFromRoot,
                     FoundFromLeft,
                     FoundFromRight };

   class WHeader;
   class WHeaderList; //instead of QList<UConfigPrivate::WHeader>
   class Record;

      // Cache class based on Header type and Record type.
   template< class RecordType > class Cache;

      // Node object and it's cache.
   class Node;
   typedef Cache< Node > NodesCache;

   class WEntry;
   typedef Cache< WEntry > ECache;

   extern const unsigned char revBitTable[];
};


      // Key, we need it here to template operation
      // inside UConfig.
   class UCKey {

        // 126 key types max.
        // Bit 8 used to determine subnode.
        // 0xFF used to determine end of subnode record.
   public:

        // Add new UCEntriesIterator constructor
        // adding new type.
      static const int knownKeyTypes;// = 2;  // ! be patient here.
      enum KeyType { IntegerKey = 0,
                     StringKey  = 1
                   };

      inline KeyType getKeyType() const { return t; }
   private:
      KeyType t;
      union {
        Q_INT32  iKey;
        QString  *sKey;
      };
   public:
      inline UCKey( const UCKey& k )
      {
        t = k.t;
        switch( t ) {
          case IntegerKey:
            iKey = k.iKey;
            break;
          case StringKey:
            sKey = new QString( *(k.sKey) );
            break;
          default:
            qFatal( "UCKey::UCKey( const UCKey &k ): unknown key type" );
        }
      }


      inline UCKey( KeyType kt ) // We will read it later
      {
        t = kt;
        switch( t ) {
          case StringKey:
            sKey = new QString;
            break;
          default:
            ;
        }
      }



      inline UCKey( const Q_INT32 k )
      {
        t = IntegerKey;

          // Our beauty hash.. How to do it faster?
        unsigned char * inAddr   = ((unsigned char * ) &k) +
                                  sizeof( k ) - 1;
        unsigned char * outAddr = (unsigned char * )  &iKey;

        *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
        *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
        *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
        *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
      }


      inline UCKey( const QString& k )
      {
        t = StringKey;
          // We need hash function here?

        sKey = new QString(k);
      }


      inline UCKey( const char* k )
      {
        t = StringKey;
          // We need hash function here?

        sKey = new QString( QString::fromLocal8Bit(k) );
      }


      inline ~UCKey()
      {
        switch( t ) {
          case StringKey:
            delete sKey;
            break;
          default:
            ;
        }
      }


      inline bool getKey( Q_INT32* v) const
      {
        if ( t == IntegerKey ) {

            // Our beauty hash.. How to do it faster?
          unsigned char * inAddr   = ((unsigned char * ) &iKey) +
	                             sizeof( iKey ) - 1;
          unsigned char * outAddr = (unsigned char * ) v;

          *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
          *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
          *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];
          *(outAddr++) = UConfigPrivate::revBitTable[ *(inAddr--) ];

          return true;
        }
      return false;
      }


      inline bool getKey( QString* v) const
      {
        if ( t == StringKey ) {
          *v = *sKey;
          return true;
        }
      return false;
      }


      inline QString getStringKey() const
      {
        #ifdef __UCDEBUG
          if ( t != StringKey )
            qFatal( "QString UCKey::getStringKey(): key type is not String" );
        #endif

        QString k;
        getKey( &k );
      return k;
      }


      inline Q_INT32 getIntegerKey() const
      {
        #ifdef __UCDEBUG
          if ( t != IntegerKey )
            qFatal( "QString UCKey::getIntegerKey(): key type is not Integer" );
        #endif

        Q_INT32 k;
        getKey( &k );
      return k;
      }


      inline QVariant getKey() const
      {
        switch( t ) {
        case IntegerKey:
          return QVariant( getIntegerKey() );
        case StringKey:
          return QVariant( getStringKey() );
        default:
          qFatal( "QVariant UCKey::getKey() : unknown key type" );
          return QVariant();
        }
      }


      inline bool operator < ( const UCKey& w ) const
      {
        if( t != w.t )
          qFatal( "bool UCKey::operator < : Could not compare keys with different types" );

        switch( t ) {
          case IntegerKey:
            return iKey < w.iKey;
          case StringKey:
            return *sKey < *w.sKey;
          default:
            qFatal( "bool UCKey::operator < : unknown key type" );
            return false;
        }
      }


      inline bool operator > ( const UCKey& w ) const
      {
        if( t != w.t )
          qFatal( "bool UCKey::operator > : Could not compare keys with different types" );

        switch( t ) {
          case IntegerKey:
            return iKey > w.iKey;
          case StringKey:
            return *sKey > *w.sKey;
          default:
            qFatal( "bool UCKey::operator > : unknown key type" );
            return false;
        }
      }


      inline Q_UINT16 checksum() const
      {
        switch( t ) {
          case IntegerKey:
            return qChecksum( (char *)&iKey, sizeof(iKey) );


          case StringKey:
            {
              QByteArray a;
              QDataStream ds( a, IO_WriteOnly );

              ds << (*sKey);

              return qChecksum( a.data(), a.size() );
            }
          default:
            qFatal( "Q_UINT16 UCKey::checksum(): unknown key type" );
          return 0;
        }
      }


      inline void load( QDataStream& ds )
      {
        switch( t ) {
          case IntegerKey:
            ds >> iKey;
            break;
          case StringKey:
            ds >> (*sKey);
            break;
          default:
            qFatal( "void UCKey::load( QDataStream& ds ): unknown key type" );
        }
      return;
      }


      inline void save( QDataStream& ds ) const
      {
        switch( t ) {
          case IntegerKey:
            ds << iKey;
            break;
          case StringKey:
            ds << (*sKey);
            break;
          default:
            qFatal( "void UCKey::save( QDataStream& ds ): unknown key type" );
        }
      }

      #if __UCDEBUG > 0
         void debug();
      #endif

      friend class UConfig;
      friend class UCEntriesIterator;
   };


class UCIterator;

class UCReadStream : public QDataStream { 
  QBuffer b;
private:
    // Stupid msvc fix :-( 
  UCReadStream( const UCReadStream& from ) :  
       QDataStream()
  {
    if( !b.setBuffer(from.b.buffer()) ) 
      qFatal( "UCReadStream( const UCReadStream & ): can't set buffer." ); 
     
    b.open( IO_ReadOnly ); 
    setDevice( &b ); 
  }


  UCReadStream( QCString from ) :  
    QDataStream( from, IO_ReadOnly ), b( from ) { }
public:
  ~UCReadStream() { }

  friend class UConfig;
  friend class UCEntriesIterator;
};


class UCWriteStream : public QDataStream {
   UConfig *c;
   QCString *data;
   UCKey k;
private:
   UCWriteStream( const UCWriteStream& from ) : QDataStream(), k( from.k ) // Never called!
   {
     qFatal( "UCWriteStream::UCWriteStream( const UCWriteStream& from ): never should be called." );
   }


   UCWriteStream( UConfig *cfg, const UCKey& key, QCString *d ) 
     : QDataStream( *d, IO_WriteOnly ),
       c(cfg), 
       data(d),
       k(key)
   {
   }

public:
   ~UCWriteStream();

   friend class UConfig;
};



/** Provide fast binary config database.
  *
  *  It looks like KConfig(KDE one), but UConfig
  *  possible to build groups in group. You can 
  *  store values under QString or int keys. Lookup 
  *  through db keys are done using b-tree.
  *  Values can have any QVariant possible type.
  *
  *  There are iterators to walk through db.
  *
  *  There are no limits of database size or value sizes.
  *  There is no limit of path depth.
  *
  *  UConfig use only Qt tools library. You can use it with 
  *  not-Qt  applications, if you need Unicoded config
  *  or local DB.
  *
  *  It can be compiled and used under Microsoft Windows.
  *  Qt tools compilings fine with Borland C++ Builder.
  *
  *
  *  Note, real saving to disk is done by flush() function.
  *  But db structure is constructing in a safe
  *  way on case of crash. Only last written values
  *  could be lose. Db structure should not be corrupted.
  *  In addition, every record header have it's own checksum.
  *
  */
class UConfig {
   static const char dbSignature[];
public:
    // FIXME, it must read root node and
    // check it for QString::null key.


    /** Encode path to provide use of special characters
     *  like '..', '/' and '.' exactly as they're look like
     *  (path "John/Jenny." is encoded to "John\/Jenny\.")
     */
   static QString encode( const QString& path );


    /** Decode path to see path with special characters
     *  exactly as it should look like (path "John\/Jenny."
     *  is decoded to "John/Jenny.").
     */
   static QString decode( const QString& path );


   enum FileType {Unknown, PlainDB, CryptedDB, EmptyDB };
    /** Checks if the file is UConfig db.
     *
     * @param fileName Checks that file
     * @return PlainDB or CryptedDB, EmptyDB if it is good db.
     *         Unknown otherwise.
     */
   static FileType isLooksLikeConfigFile( const QString & fileName );
private:
   QDataStream ds;

    // QIODevice does not have at ( 0, FROM_END )
    // operation. But file size is not changed
    // momentaly if you'll append some info to file.
    // that's why I need to know real last
    // file position.
   UConfigPrivate::FAddr lastPosition;
   UConfigPrivate::FAddr rootPosition;

   UConfigPrivate::NodesCache *nodes;
   UConfigPrivate::ECache    **eCaches;

   class LockedDir;
   LockedDir* curDir;


   QList< UCIterator > iterators;
   void iteratorDied( UCIterator * i );
   void iteratorCreated( UCIterator * i );


   UConfigPrivate::FAddr groupAddr( const bool sg );
   UConfigPrivate::FAddr getEntries( const UCKey::KeyType kt );


    // UCPathIterator call it.
   QString path( UConfigPrivate::Node* from );

   UConfigPrivate::Node* appendGroup( const UCKey& k,
                                      const int space,
                                      UConfigPrivate::WHeader *ph,
                                      const UConfigPrivate::ParentType pt );


   void reAllocCurDir();
   void correctParents( UConfigPrivate::Node *n,
                        UConfigPrivate::Node *to );

   bool internalWriteEntry( const UCKey& k,
                            const QVariant& e );

   QVariant internalReadEntry( const UCKey& k, bool *ok = 0 );
public:


   /** Construct database using specified file name.
    *
    *  It set '/' to current group and constructs new 
    *  empty database, if configFileName param
    *  point to unexistent file.
    *
    *  @param configFileName Specify database file name
    *
    *  @param fromHomeDir Specify, whether we should open file
    *                     in the user's home directory.
    *                     It is usefull like:
    *                     UConfig c( ".myprog/config.udb", true );
    *
    *                     If we have no dir '~/.myprog', UConfig
    *                     constructor will create it.
    */
   UConfig( const QString &configFileName,
            const bool fromHomeDir = false );


   /** Write all caches, detaches all iterators and close db file.
    */
   virtual ~UConfig();


   /** Set current path to read/write values there.
    *
    *  If there is no such a group setPath creates it and sets to current.
    *  '/' as the first symbol means that search starts from root.
    *  In the other way search starts from current group (spelled '.'),
    *  Symbol '..' means parent group.
    *
    *  If you want to build subgroup whith symbols like '/', "." or ".." you 
    *  should provide path in encoded form using encode().
    *
    *  Encode symbol is '\'. "\/", means '/', "\." means '.', "\\" means '\'
    *  "\<any other symbol>" comes exactly as "<any other symbol>".
    *
    *  Examples:
    *  \begin{itemize}
    *  \item setPath( "/Partners/niki" );   - from root.
    *  \item setPath( ".." );               - Set parrent group.
    *  \item setPath( "Messages" );         - from current group.
    *  \item setPath( "./Messages" );       - from current group also.
    *  \item setPath( "/User\/133" );       - Set root's subgroup "User/133".
    *  \end{itemize}
    *
    * @param p Path to set.
    * @param createNew Bool, indicates if setPath should create new
    *        subgroup if it is not exist.
    * @param space Number of subtrees pointers, which will be allocated
    *        creating new subgroups.
    * @return False on error.
    * @see path
    */
   bool setPath( const QString &p,
                 const bool createNew = true,
		 const int space = 2 );


   /** Return current path.
    *
    *  For example: '/Users/niki/Messages'
    *
    *  Path is encoded. '/' symbol encoded with "\/",
    *  '.' encoded with "\\.", '\\' encoded with "\\\\".
    *
    * @return Current path.
    * @see setPath
    */
   QString path();


   /** Flush all caches.
    *
    *  @param free true means delete cache data from memory.
    *  It can free much memory.
    *
    *  @return Number of written entries.
    */
   Q_UINT32 flush( bool free = false );


   /** Create new / update old config entry.
    *
    *  Every entry that can be stored within an QVariant can be used.
    *  Rewrite of existent entry with new type is possible.
    *  Writing QVariant::Invalide type is not possible.
    *
    *  Example:
    *  \begin{verbatim}
       UConfig c( "test.udb" );
     
       c.setPath( "/NICs/eth0" );
       c.writeEntry( "Total elements", myArraySize );
       for( int i = 0; i < myArraySize; i++ )
         c.writeEntry( i, myArray[i] );
       \end{verbatim}
    *
    *  @param k Key, may be: int or QString.
    *  @param v Value, may be any QVariant type except QVariant::Invalid.
    *  @return False on error.
    *  @see readEntry setPath
    */
   inline bool writeEntry( const UCKey& k, const QVariant& v )
   {
      return internalWriteEntry( k, v );
   }


   /** Reads bool entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       bool, false is returned and copies d into v.
     */
   bool readEntry (const UCKey &k, bool &v, bool d = false);


   /** Reads int entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       int, false is returned and copies d into v.
     */
   bool readEntry (const UCKey &k, int &v, int d = 0);


   /** Reads uint entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       uint, false is returned and copies d into v.
     */
   bool readEntry (const UCKey &k, uint &v, uint d = 0);


   /** Reads double entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       double, false is returned and copies d into v.
     */
   bool readEntry (const UCKey &k, double &v, double d = 0.0);


   /** Reads QFont entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       QFont, false is returned and copies d into v if d is not 0.
     */
   bool readEntry (const UCKey &k, QFont &v, const QFont *d = 0);


   /** Reads QPoint entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       QPoint, false is returned and copies d into v if d is not 0.
     */
   bool readEntry (const UCKey &k, QPoint &v, const QPoint *d = 0);


   /** Reads QSize entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       QSize, false is returned and copies d into v if d is not 0.
     */
   bool readEntry (const UCKey &k, QSize &v, const QSize *d = 0);


   /** Reads QRect entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       QRect, false is returned and copies d into v if d is not 0.
     */
   bool readEntry (const UCKey &k, QRect &v, const QRect *d = 0);


   /** Reads QColor entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       QColor, false is returned and copies d into v if d is not 0.
     */
   bool readEntry (const UCKey &k, QColor &v, const QColor *d = 0);


   /** Reads QString entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       QString, false is returned and copies d into v if d is not 0.
     */
   bool readEntry (const UCKey &k, QString &v, QString d = QString::null);


   /** Reads QCString entry for key k into variable v

       If no such key exists or the type of that entry cannot be casted to
       QCString, false is returned and copies d into v if d is not 0.
     */
   bool readEntry (const UCKey &k, QCString &v, const QCString *d = 0);


   /** Reads entry.
    *
    * @param k the key to the entry you are looking for.
    *          It can be an int or a QString.
    * @param ok pointer to bool value. which'll be false on reading error.
    * @return The associated QVariant value, or an
    *         invalid QVariant if  no entry with that key exists.
    */
   inline QVariant readEntry ( const UCKey &k, bool *ok = 0 )
   {
      return internalReadEntry( k, ok );
   }


   /** Deletes entry
    *
    * @param k Entry key to delete.
    * @return false on error. Error could happens if entry locked with iterator.
    */
   bool deleteEntry( const UCKey& k );


   /** Delete current node (last word in the path)
    *
    *  Root node( '/' ) can not be deleted. 
    *  Node with subnodes or entries can not be deleted.
    *  Parent node becomes current.
    *
    *  \begin{verbatim}
       UConfig c( "test.udb" );
     
       c.setPath( "/NICs/eth0" );
       c.deleteNode(); - NICs is current node now.
       c.deleteNode(); - '/' is current.
       c.deleteNode(); - nothing happend, can not delete root
       \end{verbatim}
    * @param recursively Recursively delete node with true
    * @return false on error. Error could happens if node locked with iterator
    * or contains any info inside.
    */
   bool deleteNode( const bool recursively = false );


   /** Resets statistical counters.
    */
   void resetStatistic();


   /** Statistical function
    *
    * @return Maximal number of tree iterations, during
    *         operations with QString keyed groups trees.
    */
   Q_UINT32 QStringNodesStatistic();


   /** Statistical function
    *
    * @return Maximal number of tree iterations, during
    *         operations with int keyed trees.
    */
   Q_UINT32 intTreesStatistic();


   /** Statistical function
    *
    * @return Maximal number of tree iterations, during
    *         operations with QString keyed trees.
    */
   Q_UINT32 QStringTreesStatistic();


   /**  Shows db size.
    *
    * @return Current database size.
    */
   inline Q_UINT32 streamSize() { return lastPosition; }


   /** Allows you to read own type from QCString entry stream.
    *
    * Usage: UConfig c(...); QString s; Q_INT32 i; c.readStream( "Entry key" ) >> s >> i >> ...
    *
    * @return QDataStream child class instance. Opened with IO_ReadOnly.
    * @see writeStream
    */
   inline UCReadStream readStream( const UCKey& k )
   {
     UCReadStream s( readEntry(k).toCString() ); 
     return s;
   }
 

   /** Allows you to write own type as QCString entry stream.
    *
    * Usage: UConfig c(...); c.writeStream( "Entry key" ) << QString( "string" ) << Q_INT32( -22 ) << ...
    *
    * @return QDataStream child class instance. Opened with IO_WriteOnly.
    * @see readStream
    */
   inline UCWriteStream writeStream( const UCKey& k )
   {
      return UCWriteStream( this, k, new QCString() );
   }



   friend class UCIterator;
   friend class UCPathIterator;
   friend class UCEntriesIterator;
};



  /** Base bidirectional iterator */
class UCIterator  {
public:
  virtual ~UCIterator();

  inline bool isValid() const;
  inline operator bool() { return isValid(); }

  bool toLeft();
  bool toRight();

  bool operator++();
  bool operator--();

  #ifdef __UCDEBUG
    void debugWay();
  #endif
protected:
  UConfig *c;


  UCIterator( UConfig *from,
              const UConfigPrivate::FAddr r );

  UCIterator( const UCIterator& from );


/*    UCIterator& operator = ( const UCIterator& from ) */
/*    { */
/*      // FIXME */
/*      assert( 0 ); */
/*    } */


  inline void freeWay()
  {
    while( pop() );
  return;
  }


  void detach();
  void newRoot( const UConfigPrivate::FAddr r );

  UConfigPrivate::WHeader* top();
private:
  UConfigPrivate::WHeaderList *headerStack;

  UConfigPrivate::FAddr root;

  UConfigPrivate::WHeader* pop();
  void push( UConfigPrivate::WHeader *h );

  virtual UConfigPrivate::WHeader* getHeader( UConfigPrivate::FAddr a ) = 0;

  friend class UConfig; // for detaching..
};




  /** Walking through groups iterator.
   * It walks through groups of const sublevel.
   * It means we do not go into subgroups.
   */

class UCPathIterator : public UCIterator {
  virtual UConfigPrivate::WHeader* getHeader( UConfigPrivate::FAddr a );
public:
  UCPathIterator( UConfig *from,
                  const bool subGroupIterator = true );

  UCPathIterator( const UCPathIterator& from );

  UCPathIterator( const UCPathIterator& from,
                  const bool subGroupIterator );

  ~UCPathIterator();


    // This operations do not affect on curDir.
  QString path(); // Return full path.
                  // QString::null if iterator is invalid.


  QString group(); // Return name of current group.


  bool setPath();  // Sets path in uconfig to path
                   // of this iterator. false, if something
                   // wrong.
};



  /** Walking through entries iterator.
   * It walks on entries of specified type.
   * Type is specified during construction.
   */

class UCEntriesIterator : public UCIterator {
  virtual UConfigPrivate::WHeader* getHeader( UConfigPrivate::FAddr a );

  UCKey::KeyType kt;

public:

  UCEntriesIterator( UConfig *c, const int );
  UCEntriesIterator( UConfig *c, const QString& );

  // FIXME do it. ;)
//  template< class KeyT >
//  UCEntriesIterator( const UCPathIterator& from, const KeyT& kv )
//    : UCIterator( from )
//  {
//    UCKey k( kv );
//    kt = k.getKeyType();
//    toLeft();
//  }

  UCEntriesIterator( const UCEntriesIterator& from );

  ~UCEntriesIterator();

  QVariant readEntry( bool *ok = 0 );


   /** Allows you to read own type from QCString entry stream.
    *
    * @return QDataStream child class instance. Opened with IO_ReadOnly.
    * @see writeStream
    */
   inline UCReadStream readStream()
   {
     UCReadStream s( readEntry().toCString() ); 
     return s;
   }

  QVariant getKey();

//  template< class ValueT >
//  inline bool readEntry( ValueT* v )
//  {
//    return readEntry(hack).getValue( v );
//  }


  inline QVariant::Type entryType()
  {
    return readEntry().type();
  }
};


#endif
